use std::{io, str::Utf8Error};
use thiserror::Error;

#[derive(Error, Debug)]
pub enum Error {
    #[error("IO Error: {0}")]
    Io(#[from] io::Error),
    #[error("Parse utf8 data: {0}")]
    Utf8(#[from] Utf8Error),
    #[error("Simplification of Element Tree results in no Elements")]
    Simplify,
}
