use super::Element;
use serde::{Deserialize, Serialize};

#[derive(Debug, Clone, Deserialize, PartialEq, Eq, Serialize)]
pub struct ListItem {
    pub bullet: ListBullet,
    pub element: Element,
}

#[derive(Debug, Clone, Deserialize, PartialEq, Eq, Serialize)]
pub enum ListBullet {
    None,
    Point,
    Number(i32),
}

#[derive(Debug, Clone, Deserialize, PartialEq, Eq, Serialize)]
pub enum ListType {
    None,
    Bullet,
    Numbered { start: i32 },
}
