#[derive(Debug, Default, Clone, Copy, PartialEq, Eq, glib::Enum)]
#[enum_type(name = "SelfStackTransition")]
pub enum SelfStackTransition {
    #[default]
    Crossfade,
    SlideLeft,
    SlideRight,
}
