use glib::Enum;
use serde::{Deserialize, Serialize};

#[derive(Clone, Copy, Default, Debug, Serialize, PartialEq, Deserialize, Enum)]
#[repr(u32)]
#[enum_type(name = "ArticleListMode")]
pub enum ArticleListMode {
    #[default]
    All,
    Unread,
    Marked,
}
