// This file was generated by gir (https://github.com/gtk-rs/gir)
// from 
// from ../gir-files-gstreamer
// from ../gir-files-gtk
// DO NOT EDIT

use crate::{ffi};
use glib::{GStr};

#[doc(alias = "CLAPPER_TIME_FORMAT")]
pub static TIME_FORMAT: &GStr = unsafe{GStr::from_utf8_with_nul_unchecked(ffi::CLAPPER_TIME_FORMAT)};
#[doc(alias = "CLAPPER_TIME_MS_FORMAT")]
pub static TIME_MS_FORMAT: &GStr = unsafe{GStr::from_utf8_with_nul_unchecked(ffi::CLAPPER_TIME_MS_FORMAT)};
#[doc(alias = "CLAPPER_VERSION_S")]
pub static VERSION_S: &GStr = unsafe{GStr::from_utf8_with_nul_unchecked(ffi::CLAPPER_VERSION_S)};
