
    #[derive(Debug)]
    pub struct PasswordCryptKey;

    impl PasswordCryptKey {
        pub fn get() -> String {
            obfstr::obfstr!("ypsSXCLhJngks9qGUAqShd8JjRaZ824wT3e").into()
        }
    }
    